﻿using System;
using System.Collections.Generic;
using System.Linq;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRageMath;

namespace Scripts.KSWH
{
    class BattleShip : Ai
    {
        public const int Price = 500;
        public static string HudAttackMessage = Hud.MessageLead + "///WARNING::                                                                                                  \n" +
                                                 ">Incomming signature detected\n    -SIZE: Big\n    -ARMOR: Heavy\n    -OFFENSE: Heavy\n>HIGH THREAT LEVEL\n\n***ATTACK IMMINENT***";

        public BattleShip(IMyCubeGrid grid, int agroRange, int id)
        {
            MaxCount = 50;
            SetId(id);
            Grid = grid;
            Range = agroRange;
            //m_cockpit = controller;
            IsControlable = true;
            SetOwsnershipToEnemy(grid);
            FindTarget();
        }

        public BattleShip(IMyCubeGrid grid, int agroRange, IMyControllableEntity controller)
        {
            MaxCount = 50;
            Grid = grid;
            Range = agroRange;
            m_cockpit = controller;
            IsControlable = (controller != null);
            SetOwsnershipToEnemy(grid);
            if (IsControlable)
            {
                FindTarget();
            }
        }

        public override int GetBounty()
        {
            return 20;
        }

        public override void Update()
        {
            Count++;

            if (Count == 100) //every 100th time the computer passes here it should trigger the localization script
            {
                FindTarget(); //simple hold so we are not overloading the cpu
                Count = 0;
            }

            base.Update();

            if (!IsControlable)
            {
                return;
            }

            if (Target == null)
            {
                return; //no target to go to
            }

            if (!IsSpotted)
            {
                IsSpotted = true;
                MissionComponent.AlarmHandler.NewSig();
            }

            var dir = Target.GetPosition() - Grid.GetPosition();

            if (!IsAttacking && dir.Length() < 5000)
            {
                IsAttacking = true;
                MissionComponent.AlarmHandler.Alert((int)Hud.MessagePriority.Battleship, HudAttackMessage);
            }

            if (dir.Length() > 1000)
            {
               if (nav != null)
                    {
                        if (nav.CNS.AutoPilotReady)
                        {
                            var loc = GetTargetDestination(Target.GetPosition(), -400, 400);
                            if (!loc.HasValue)
                            {
                                return;
                            }

                            
                            nav.CNS.setDestination(loc);
                        }
                        return;
                    }
                    var locb = GetTargetDestination(Target.GetPosition(), -400, 400);
                    if (!locb.HasValue)
                    {
                        MissionComponent.Speak("ERROR", "nav cannot locate waypoint");
                        return;
                    }

                    nav = MissionComponent.AutoHandler.AddNavigator(Grid,locb);
                    return; //outside of aggro range
            }

            if (nav != null)
            {
                MissionComponent.AutoHandler.RemoveNavigator(Grid);
                nav = null;
            }

            var dirNorm = Vector3.Normalize(dir);
            var x = (float)-(m_cockpit as IMyEntity).WorldMatrix.Up.Dot(dirNorm);
            var y = (float)-(m_cockpit as IMyEntity).WorldMatrix.Left.Dot(dirNorm);
            var forw = (m_cockpit as IMyEntity).WorldMatrix.Forward.Dot(dirNorm);

            if (forw < 0)
                y = 0;
            if (Math.Abs(x) < 0.2f)
                x = 0;
            if (Math.Abs(y) < 0.2f)
                y = 0;

            

            if (dir.Length() < 500 && !SelfDestructing)
            {
                dir = Vector3.Zero;
            }
            else if (dir.Length() < 1000 && !SelfDestructing) 
            {
                //break;
                var getMatrix = (Matrix)(m_cockpit as IMyEntity).WorldMatrixNormalizedInv;
                Matrix newDirection = new Matrix();
                newDirection.Backward = (getMatrix.Backward / 2);
                newDirection.Forward = (getMatrix.Forward / 2);
                newDirection.Down = (getMatrix.Down / 2);
                newDirection.Up = (getMatrix.Up / 2);
                newDirection.Left = (getMatrix.Left / 2);
                newDirection.Right = (getMatrix.Right / 2);
                dir = Vector3.TransformNormal(dir, newDirection);
            }
            else
            {
                //strom to target
                dir = Vector3.TransformNormal(dir, (Matrix)(m_cockpit as IMyEntity).WorldMatrixNormalizedInv);
            }

            var rot = new Vector2(x, y) * 10f;
            if (dir.LengthSquared() > 0 || rot.LengthSquared() > 0)
            {
                m_cockpit.MoveAndRotate(dir, rot, 0);
            }
            else
            {
                //MissionComponent.Speak("AI_overwrite", "No movement actions");
                m_cockpit.MoveAndRotateStopped();
            }
            return;
        }

        protected override void UseEcm()
        {
            if (!MissionComponent.StationHandler.EcmReady)
            {
                return; //incursion state 6 not reached
            }

            if (UsedEcm)
            {
                return; //power already used
            }

            for (int i = 0; i < 10; i++)
            {
                MissionComponent.EcmHandler.LaunchMetoer(Target.GetPosition(), Grid.GetPosition(), Grid.WorldMatrix, EcmActions.Size.Small);
            }

            for (int i = 0; i < 4; i++)
            {
                MissionComponent.EcmHandler.LaunchMetoer(Target.GetPosition(), Grid.GetPosition(), Grid.WorldMatrix, EcmActions.Size.Large);
            }
            UsedEcm = true;
        }

        protected override int GetLivespawn()
        {
            return 30;
        }

        private void FindTarget()
        {
            ThreadWrapper.TargetLock.AcquireExclusive();
            try
            {
                var bs = new BoundingSphere(Grid.GetPosition(), Range);
                var sphere = new BoundingSphereD(Grid.GetPosition(), Range*1.5);
                var ents = MyAPIGateway.Entities.GetEntitiesInSphere(ref sphere);
                if (ents.Count > 0)
                {
                    MyDefinitionId decoyDef = new MyDefinitionId(typeof (MyObjectBuilder_Decoy));
                    foreach (IMyEntity ent in ents)
                    {
                        var grid = ent as IMyCubeGrid;
                        if (grid == null) continue;
                        var lst = new List<IMySlimBlock>();
                        grid.GetBlocks(lst,
                            x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeId.Equals(decoyDef.TypeId));
                        if (lst.Count <= 0) continue;
                        Target = grid;
                        ThreadWrapper.TargetLock.ReleaseExclusive();
                        return;
                    }
                }


                Target = GetPlayerInSphere(ref bs) ?? MissionComponent.firewall.FatBlock.CubeGrid;
                if (Target != null)
                {
                    DisableOverride();
                }
            }
            catch (Exception)
            {
                //ignored
            }
            
            ThreadWrapper.TargetLock.ReleaseExclusive();
        }
    }
}
