﻿using System;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRageMath;

namespace Scripts.KSWH
{
    class BattleshipStation : Station
    {
        private const int MinDistance = 12;
        private const int ProductionTime = 20;
        private const int ResourceBonus = 200;
        public const int BuildCosts = 950;

        public BattleshipStation()
        {
            Random r = new Random();
            Type = StationType.Battleship;
            int x = (r.Next(20) + MinDistance) * 1000 + MissionComponent.firewall.Position.X;
            int y = (r.Next(20) + MinDistance) * 1000 + MissionComponent.firewall.Position.Y;
            int z = (r.Next(20) + MinDistance) * 1000 + MissionComponent.firewall.Position.X;
            ThreadWrapper.OrderSpawn(StaticInformation.FrigShipyard[r.Next(StaticInformation.FrigShipyard.Length - 1)], new Vector3D(x, y, z), this);
            Income = ResourceBonus;
        }

        public BattleshipStation(IMyCubeGrid grid)
        {
            Grid = grid;
            Type = StationType.Battleship;
            Income = ResourceBonus;
            SetOwsnershipToEnemy();
        }


        public void StartProduction()
        {
            if (!IsProducing)
            {
                if (!MissionComponent.host.AiHasMoney(BattleShip.Price))
                {
                    return;
                }
                Random r = new Random(DateTime.Now.Minute);
                string prefab = StaticInformation.Frigates[r.Next(StaticInformation.BattleshipPrefabList.Length - 1)];
                base.StartProduction(ProductionTime, prefab);
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Battleship station", "Building battleship");
                }
            }
        }

        public static void HookAi(IMyEntity ent, int id)
        {
            BattleShip b = new BattleShip(ent as IMyCubeGrid, StaticInformation.AggroRange, id);
            MissionComponent.AddAiShip(b, id);
        }

        public override int GetBounty()
        {
            return 60;
        }
    }
}
