﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using Scripts.KSWH;
using VRageMath;

namespace Scripts.KSWH
{
    class Compass
    {
        public IMyCubeGrid Grid { private set; get; }
        
        private IMyControllableEntity m_cockpit;
        public bool isControllable { private set; get; }

        public Compass(IMyCubeGrid grid)
        {
            Grid = grid;
            StillControlable();
            if (MissionComponent.writeDebug)
            {
                MissionComponent.Speak("alarm", "New compas located, controllable " + isControllable);
            }

        }

        /// <summary>
        /// Give the compas calculation time so it can point at a target
        /// </summary>
        public void Update(Vector3D pointTarget)
        {

            if (!StillControlable())
            {
                return; //ai needs to remove the hook
            }
            PointToTarget(pointTarget);
        }

        /// <summary>
        /// Check if the compast is still valid (has a controllable entity)
        /// </summary>
        /// <returns>boolean if the ship is controllable</returns>
        private bool StillControlable()
        {
            var tmp = new List<IMySlimBlock>();

            // Get a list of cockpit/remote control blocks
            
            Grid.GetBlocks(tmp,
                s => s.FatBlock != null
                     && (s.FatBlock.BlockDefinition.TypeIdString == "MyObjectBuilder_Cockpit"
                         || s.FatBlock.BlockDefinition.TypeIdString == "MyObjectBuilder_RemoteControl"
                         )
                );

            var lst = (from block in tmp
                       let definition = MyDefinitionManager.Static.GetCubeBlockDefinition(block.FatBlock.BlockDefinition)
                       where
                           definition is MyShipControllerDefinition &&
                           (definition as MyShipControllerDefinition).EnableShipControl
                       select block).ToList();


            if (lst.Count > 0)
            {
                m_cockpit = lst[0].FatBlock as IMyControllableEntity;
                isControllable = true;
                return true;
            }
            isControllable = false;
            return false;
            
        }

        /// <summary>
        /// Get the target the compas should 'point' towards and point to it
        /// </summary>
        private void PointToTarget(Vector3D stationLocation)
        {
            
            var dirNorm = Vector3.Normalize(stationLocation);
            var x = (float)-(m_cockpit as IMyEntity).WorldMatrix.Up.Dot(dirNorm);
            var y = (float)-(m_cockpit as IMyEntity).WorldMatrix.Left.Dot(dirNorm);
            var forw = (m_cockpit as IMyEntity).WorldMatrix.Forward.Dot(dirNorm);

            if (forw < 0)
                y = 0;
            if (Math.Abs(x) < 0.2f)
                x = 0;
            if (Math.Abs(y) < 0.2f)
                y = 0;
            var rot = new Vector2(x, y) * 10f;
            if (rot.LengthSquared() > 0)
            {
                m_cockpit.MoveAndRotate(new Vector3(), rot, 0);
            }
            else
            {
                m_cockpit.MoveAndRotateStopped();
            }
        }
    }
}
