﻿using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using System;
using System.Collections.Generic;
using System.Linq;
using Sandbox.Common;
using Sandbox.Definitions;
using VRageMath;

namespace Scripts.KSWH
{
    class Cruiser : Ai
    {
        //public readonly int Range = 10000;
        public static int Price = 300;
        
        private bool usedEmp;
        private int rangeCount;
        private readonly int abilityRange = 1000;
        public static string HudAttackMessage = Hud.MessageLead + "///WARNING::                                                                                                  \n" +
                                                 ">Incomming signature detected\n    -SIZE: Average\n    -ARMOR: Average\n    -OFFENSE: Heavy\n>MEDIUM THREAT LEVEL\n\n***ATTACK IMMINENT***";

        public Cruiser(IMyCubeGrid grid, int agroRange, int id)
        {
            MaxCount = 100;
            SetId(id);
            Grid = grid;
            Range = agroRange;
            //m_cockpit = controller;
            IsControlable = true;
            usedEmp = false;
            rangeCount = 0;
            SetOwsnershipToEnemy(grid);
        }

        public Cruiser(IMyCubeGrid grid, int agroRange, IMyControllableEntity controller)
        {
            MaxCount = 100;
            Grid = grid;
            Range = agroRange;
            m_cockpit = controller;
            IsControlable = (controller != null);
            usedEmp = false;
            rangeCount = 0;
            SetOwsnershipToEnemy(grid);
        }


        public override int GetBounty()
        {
            return 10;
        }

        public override void Update() 
        {
            //todo: ugly/lazy way of fixing the nullpointer when building a new Grid during a update cycle, should be fixed properly
            Count++;
            try
            {
                if (Count == 100) //every 100th time the computer passes here it should trigger the localization script
                {
                    FindTarget(); //simple hold so we are not overloading the cpu
                    Count = 0;
                }

                base.Update();

                if (!IsControlable)
                {
                    return;
                }

                if (Target == null)
                {
                    return; //no target to go to
                }

                if (!IsSpotted)
                {
                    IsSpotted = true;
                    MissionComponent.AlarmHandler.NewSig();
                }

                var dir = Target.GetPosition() - Grid.GetPosition();

                if (!IsAttacking && dir.Length() < 5000)
                {
                    IsAttacking = true;
                    MissionComponent.AlarmHandler.Alert((int)Hud.MessagePriority.Cruiser, HudAttackMessage);
                }

                if (dir.Length() > 1000)
                {
                    if (nav != null)
                    {
                        if (nav.CNS.AutoPilotReady)
                        {
                            var loc = GetTargetDestination(Target.GetPosition(), -200, 200);
                            if (!loc.HasValue)
                            {
                                return;
                            }

                            
                            nav.CNS.setDestination(loc);
                        }
                        return;
                    }
                    var locb = GetTargetDestination(Target.GetPosition(), -200, 200);
                    if (!locb.HasValue)
                    {
                        MissionComponent.Speak("ERROR", "nav cannot locate waypoint");
                        return;
                    }

                    nav = MissionComponent.AutoHandler.AddNavigator(Grid,locb);
                    return; 
                }

                if (nav != null)
                {
                    MissionComponent.AutoHandler.RemoveNavigator(Grid);
                    nav = null;
                }
                
                
                var dirNorm = Vector3.Normalize(dir);
                var x = (float) -(m_cockpit as IMyEntity).WorldMatrix.Up.Dot(dirNorm);
                var y = (float) -(m_cockpit as IMyEntity).WorldMatrix.Left.Dot(dirNorm);
                var forw = (m_cockpit as IMyEntity).WorldMatrix.Forward.Dot(dirNorm);

                if (forw < 0)
                    y = 0;
                if (Math.Abs(x) < 0.2f)
                    x = 0;
                if (Math.Abs(y) < 0.2f)
                    y = 0;
                //dir = dir.Length() < 100 ? Vector3.Zero : Vector3.TransformNormal(dir, (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv);
                

                if (dir.Length() < 500 && !SelfDestructing)
                {
                    dir = Vector3.Zero;
                } else if (dir.Length() < 750 && !SelfDestructing)
                {
                    //break;
                    var getMatrix = (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv;
                    Matrix newDirection = new Matrix();
                    newDirection.Backward = (getMatrix.Backward/3);
                    newDirection.Forward = (getMatrix.Forward/3);
                    newDirection.Down = (getMatrix.Down/3);
                    newDirection.Up = (getMatrix.Up/3);
                    newDirection.Left = (getMatrix.Left/3);
                    newDirection.Right = (getMatrix.Right/3);
                    dir = Vector3.TransformNormal(dir,newDirection);
                }
                else
                {
                    //strom to target
                    dir = Vector3.TransformNormal(dir, (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv);
                }

                var rot = new Vector2(x, y)*10f;
                if (dir.LengthSquared() > 0 || rot.LengthSquared() > 0)
                {
                    m_cockpit.MoveAndRotate(dir, rot, 0);
                }
                else
                {
                    //MissionComponent.Speak("AI_overwrite", "No movement actions");
                    m_cockpit.MoveAndRotateStopped();
                }
                    
            }
            catch (Exception e)
            {
                //MissionComponent.Speak("System", "Error: " + e);
                
            }
        }

        protected override void UseEcm()
        {
            if (!MissionComponent.StationHandler.EcmReady)
            {
                return;//not in incursion state 6
            }

            if (UsedEcm)
            {
                return;//power already used
            }

            MissionComponent.EcmHandler.EmpBlast(new BoundingSphereD(Grid.GetPosition(), 5000));
            UsedEcm = true;
        }

        protected override int GetLivespawn()
        {
            return 20;
        }

        private void FindTarget()
        {
            ThreadWrapper.TargetLock.AcquireExclusive();
            try
            {
                var bs = new BoundingSphere(Grid.GetPosition(), Range);
                Target = GetPlayerInSphere(ref bs) ?? MissionComponent.firewall.FatBlock.CubeGrid;
                if (Target != null)
                {
                    DisableOverride();
                }
            }
            catch (Exception)
            {
                //ignored
            }
            
            ThreadWrapper.TargetLock.ReleaseExclusive();
        }

    }
}
