﻿using System;
using System.Collections.Generic;
using System.Linq;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRageMath;

namespace Scripts.KSWH
{
    class Destroyer : Ai
    {
        public static int Price = 100;
        public static string HudAttackMessage = Hud.MessageLead + "///WARNING::                                                                                                  \n" +
                                                 ">Incomming signature detected\n    -SIZE: small\n    -ARMOR: low\n    -OFFENSE: Average\n>LOW THREAT LEVEL\n\n***ATTACK IMMINENT***";

        public Destroyer(IMyCubeGrid grid, int agroRange, int id)
        {
            MaxCount = 100;
            SetId(id);
            Grid = grid;
            Range = agroRange;
            //m_cockpit = controller;
            IsControlable = true;
            SetOwsnershipToEnemy(grid);
        }

        public Destroyer(IMyCubeGrid grid, int agroRange, IMyControllableEntity controller)
        {
            MaxCount = 100;
            Grid = grid;
            Range = agroRange;
            m_cockpit = controller;
            IsControlable = (controller != null);
            SetOwsnershipToEnemy(grid);
        }

        public override int GetBounty()
        {
            return 5;
        }

        public override void Update()
        {
            Count++;

            if (Count == 100) //every 100th time the computer passes here it should trigger the localization script
            {
                FindTarget(); //simple hold so we are not overloading the cpu
                Count = 0;
            }

            base.Update();

            if (!IsControlable)
            {
                return;
            }

            if (Target == null)
            {
                return; //no target to go to
            }

            if (!IsSpotted)
            {
                IsSpotted = true;
                MissionComponent.AlarmHandler.NewSig();
            }

            var dir = Target.GetPosition() - Grid.GetPosition();

            if (!IsAttacking && dir.Length() < 5000)
            {
                IsAttacking = true;
                MissionComponent.AlarmHandler.Alert((int)Hud.MessagePriority.Destroyer, HudAttackMessage);
            }

            if (dir.Length() > 500)
            {
                if (nav != null)
                {
                    if (nav.CNS.AutoPilotReady)
                    {
                        var loc = GetTargetDestination(Target.GetPosition(), -200, 200);
                        if (!loc.HasValue)
                        {
                            return;
                        }


                        nav.CNS.setDestination(loc);
                    }
                    return;
                }
                var locb = GetTargetDestination(Target.GetPosition(), -200, 200);
                if (!locb.HasValue)
                {
                    MissionComponent.Speak("ERROR", "nav cannot locate waypoint");
                    return;
                }

                nav = MissionComponent.AutoHandler.AddNavigator(Grid, locb);
                return; 
            }

            if (nav != null)
            {
                MissionComponent.AutoHandler.RemoveNavigator(Grid);
                nav = null;
            }
            

            var dirNorm = Vector3.Normalize(dir);
            var x = (float) -(m_cockpit as IMyEntity).WorldMatrix.Up.Dot(dirNorm);
            var y = (float) -(m_cockpit as IMyEntity).WorldMatrix.Left.Dot(dirNorm);
            var forw = (m_cockpit as IMyEntity).WorldMatrix.Forward.Dot(dirNorm);

            if (forw < 0)
                y = 0;
            if (Math.Abs(x) < 0.2f)
                x = 0;
            if (Math.Abs(y) < 0.2f)
                y = 0;

            //dir = dir.Length() < 100 ? Vector3.Zero : Vector3.TransformNormal(dir, (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv);
            if (dir.Length() < 1000)
            {
                UseEcm();
            }

            if (dir.Length() < 300 && !SelfDestructing)
            {
                dir = Vector3.Zero;
            }
            else if (dir.Length() < 400 && !SelfDestructing)
            {
                //break;
                var getMatrix = (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv;
                Matrix newDirection = new Matrix();
                newDirection.Backward = (getMatrix.Backward/3);
                newDirection.Forward = (getMatrix.Forward/3);
                newDirection.Down = (getMatrix.Down/3);
                newDirection.Up = (getMatrix.Up/3);
                newDirection.Left = (getMatrix.Left/3);
                newDirection.Right = (getMatrix.Right/3);
                dir = Vector3.TransformNormal(dir, newDirection);
            }
            else
            {
                //strom to target
                dir = Vector3.TransformNormal(dir, (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv);
            }

            var rot = new Vector2(x, y)*10f;
            if (dir.LengthSquared() > 0 || rot.LengthSquared() > 0)
            {
               m_cockpit.MoveAndRotate(dir, rot, 0);
            }
            else
            {
                //MissionComponent.Speak("AI_overwrite", "No movement actions");
                m_cockpit.MoveAndRotateStopped();
            }
            return;
        }

        protected override void UseEcm()
        {
            if (!MissionComponent.StationHandler.EcmReady)
            {
                return; //incursion state 6 aint there
            }

            if (UsedEcm)
            {
                return; //unit already used power
            }
            MissionComponent.EcmHandler.ScrapStorm(EcmActions.Size.Small, Target.GetPosition());
            UsedEcm = true;
        }

        protected override int GetLivespawn()
        {
            return 15;
        }

        private void FindTarget()
        {
            ThreadWrapper.TargetLock.AcquireExclusive();
            try
            {
                var bs = new BoundingSphere(Grid.GetPosition(), Range);
                Target = GetPlayerInSphere(ref bs) ?? MissionComponent.firewall.FatBlock.CubeGrid;
                if (Target != null)
                {
                    DisableOverride();
                }
            }
            catch (Exception)
            {
                //ingored
            }
            finally
            {
                ThreadWrapper.TargetLock.ReleaseExclusive();
            }
            
            
        }

    }
}
