﻿using System;
using System.Collections.Generic;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRageMath;

namespace Scripts.KSWH
{
    class Frigate : Ai
    {
        public static int Price = 50;
        private bool mFiringGuns;
        private static readonly string HudAttackMessage = Hud.MessageLead + "///WARNING::                                                                                                  \n" +
                                                 ">Incomming signature detected\n    -SIZE: small\n    -ARMOR: low\n    -OFFENSE: low\n>LOW THREAT LEVEL\n\n***ATTACK IMMINENT***";

        public Frigate(IMyCubeGrid grid, int agroRange, int id)
        {
            MaxCount = 50;
            SetId(id);
            Grid = grid;
            Range = agroRange;
            IsControlable = true;
            SetOwsnershipToEnemy(grid);
            FindTarget();
        }

        public Frigate(IMyCubeGrid grid, int agroRange, IMyControllableEntity controller)
        {
            MaxCount = 50;
            Grid = grid;
            Range = agroRange;
            m_cockpit = controller;
            IsControlable = (controller != null);
            SetOwsnershipToEnemy(grid);
        }

        public override int GetBounty()
        {
            return 5;
        }

        public override void Update()
        {
            Count++;
            if (Count == 100) //every 100th time the computer passes here it should trigger the localization script
            {
                FindTarget(); //simple hold so we are not overloading the cpu
                Count = 0;
            }

            base.Update();
            if (!IsControlable)
            {
                return;
            }

            if (Target == null)
            {
                FindTarget();
            }

            if (!IsSpotted)
            {
                IsSpotted = true;
                MissionComponent.AlarmHandler.NewSig();
            }

            var dir = Target.GetPosition() - Grid.GetPosition();

            
            if (!IsAttacking && dir.Length() < 5000)
            {
                IsAttacking = true;
                MissionComponent.AlarmHandler.Alert((int)Hud.MessagePriority.Frigate, HudAttackMessage);
            }
            //check if guns should be firing
            if (dir.Length() < 600)
            {
                if (!mFiringGuns)
                {
                    mFiringGuns = FireMountedGunsStart(true);
                }
                UseEcm();
            }
            else
            {
                if (mFiringGuns)
                {
                    mFiringGuns = FireMountedGunsStart(false);
                }
            }

            if (dir.Length() > 500)
            {
                if (nav != null)
                {
                    if (nav.CNS.AutoPilotReady)
                    {
                        var loc = GetTargetDestination(Target.GetPosition(), -200, 200);
                        if (!loc.HasValue)
                        {
                            return;
                        }


                        nav.CNS.setDestination(loc);
                    }
                    return;
                }
                var locb = GetTargetDestination(Target.GetPosition(), -200, 200);
                if (!locb.HasValue)
                {
                    MissionComponent.Speak("ERROR", "nav cannot locate waypoint");
                    return;
                }

                nav = MissionComponent.AutoHandler.AddNavigator(Grid, locb);
                return; //outside of aggro range
            }

            if (nav != null)
            {
                MissionComponent.AutoHandler.RemoveNavigator(Grid);
                nav = null;
            }

            var dirNorm = Vector3.Normalize(dir);
            var x = (float) -(m_cockpit as IMyEntity).WorldMatrix.Up.Dot(dirNorm);
            var y = (float) -(m_cockpit as IMyEntity).WorldMatrix.Left.Dot(dirNorm);
            var forw = (m_cockpit as IMyEntity).WorldMatrix.Forward.Dot(dirNorm);

            if (forw < 0)
                y = 0;
            if (Math.Abs(x) < 0.2f)
                x = 0;
            if (Math.Abs(y) < 0.2f)
                y = 0;
            

            if (dir.Length() < 100 && !SelfDestructing)
            {
                dir = Vector3.Zero;
            }
            else if (dir.Length() < 250 && !SelfDestructing)
            {
                //break;
                var getMatrix = (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv;
                Matrix newDirection = new Matrix();
                newDirection.Backward = (getMatrix.Backward/2);
                newDirection.Forward = (getMatrix.Forward/2);
                newDirection.Down = (getMatrix.Down/2);
                newDirection.Up = (getMatrix.Up/2);
                newDirection.Left = (getMatrix.Left/2);
                newDirection.Right = (getMatrix.Right/2);
                dir = Vector3.TransformNormal(dir, newDirection);
            }
            else
            {
                //strom to target
                dir = Vector3.TransformNormal(dir, (Matrix) (m_cockpit as IMyEntity).WorldMatrixNormalizedInv);
            }

            var rot = new Vector2(x, y)*20f; //*10f org 
            if (dir.LengthSquared() > 100 || rot.LengthSquared() > 0)
            {
                m_cockpit.MoveAndRotate(dir, rot, 0);
            }
            else
            {
                m_cockpit.MoveAndRotateStopped();
            }
        }

        protected override void UseEcm()
        {
            if (!MissionComponent.StationHandler.EcmReady)
            {
                return; //incurstion state 6 not active
            }
            if (UsedEcm)
            {
                return; //allready spend power with this unit
            }
            //fire 3 ' bomms' 
            for (int i = 0; i < 3; i++)
            {
                MissionComponent.EcmHandler.LaunchMetoer(Target.GetPosition(), Grid.GetPosition(), Grid.WorldMatrix, EcmActions.Size.Small);
            }
            UsedEcm = true;
        }

        protected override int GetLivespawn()
        {
            return 10;
        }

        private void FindTarget()
        {
            ThreadWrapper.TargetLock.AcquireExclusive();
            try
            {
                var bs = new BoundingSphere(Grid.GetPosition(), Range);
                var sphere = new BoundingSphereD(Grid.GetPosition(), Range/2);

                var ents = MyAPIGateway.Entities.GetEntitiesInSphere(ref sphere);
                if (ents.Count > 0)
                {
                    MyDefinitionId decoyDef = new MyDefinitionId(typeof (MyObjectBuilder_Decoy));
                    foreach (IMyEntity ent in ents)
                    {
                        var grid = ent as IMyCubeGrid;
                        if (grid == null) continue;
                        var lst = new List<IMySlimBlock>();
                        grid.GetBlocks(lst,
                            x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeId.Equals(decoyDef.TypeId));
                        if (lst.Count <= 0) continue;
                        Target = grid;
                        ThreadWrapper.TargetLock.ReleaseExclusive();
                        return;
                    }
                }



                Target = GetPlayerInSphere(ref bs) ?? MissionComponent.firewall.FatBlock.CubeGrid;
                if (Target != null)
                {
                    DisableOverride();
                }
            }
            catch (Exception)
            {
                // ignored
            }
            finally
            {
                ThreadWrapper.TargetLock.ReleaseExclusive(); 
            }

            
        }
    }
}
