﻿using System;
using Sandbox.ModAPI;
using VRageMath;

namespace Scripts.KSWH
{
    class FrigateStation : Station
    {
        private const int MinDistance = 5;
        private const int ProductionTime = 2;
        private const int ResourceBonus = 75;
        public const int BuildCosts = 200;

        public FrigateStation()
        {
            Random r = new Random();
            Type = StationType.Frigate;
            ThreadWrapper.OrderSpawn(StaticInformation.FrigShipyard[r.Next(StaticInformation.FrigShipyard.Length - 1)], GenerateLocation(MinDistance), this);
            Income = ResourceBonus;
            SetOwsnershipToEnemy();
        }

        public FrigateStation(IMyCubeGrid grid)
        {
            Grid = grid;
            Income = ResourceBonus;
            Type = StationType.Frigate;
            SetOwsnershipToEnemy();
        }

        public void StartProduction()
        {
            if (!IsProducing)
            {
                if (!MissionComponent.host.AiPay(Frigate.Price))
                {
                    return;
                }
                Random r = new Random(DateTime.Now.Minute);
                string prefab = StaticInformation.Frigates[r.Next(StaticInformation.Frigates.Length - 1)];
                base.StartProduction(ProductionTime, prefab);
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Frigate station", "Now producing new frigate");
                }
            }
            
        }

        public static void HookAi(IMyEntity ent, int id)
        {
            Frigate f = new Frigate(ent as IMyCubeGrid, StaticInformation.AggroRange, id);
            MissionComponent.AddAiShip(f, id);
        }

        public override int GetBounty()
        {
            return 15;
        }
    }
}
