﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sandbox.ModAPI;
using Scripts.KSWH;
using VRageMath;

namespace Scripts.KSWH
{
    class LargeLaunchPlatform : Station
    {
        private const int MinDistance = 15;
        private const int ProductionTime = 10;
        private const int ResourceBonus = 40;
        public static int BuildCost = 600;

        public LargeLaunchPlatform()
        {
            Random r = new Random();
            Type = StationType.LargeLauncher;
            ThreadWrapper.OrderSpawn(
                StaticInformation.LargeLaunchPlatforms[r.Next(StaticInformation.LargeLaunchPlatforms.Length - 1)],
                GenerateLocation(MinDistance), this);
            Income = ResourceBonus;
            SetOwsnershipToEnemy();
        }

        public LargeLaunchPlatform(IMyCubeGrid grid)
        {
            Grid = grid;
            Type = StationType.LargeLauncher;
            Income = ResourceBonus;
            SetOwsnershipToEnemy();
        }

        public void StartProduction()
        {
            if (!IsProducing)
            {
                if (!MissionComponent.host.AiPay(750))
                {
                    return;
                }
                Random r = new Random(DateTime.Now.Minute);
                string prefab =
                    StaticInformation.LargeBallisticWeapons[r.Next(StaticInformation.LargeBallisticWeapons.Length - 1)];
                base.StartProduction(ProductionTime, prefab);
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Large Launcher", "Now producing missile");
                }
            }
        }

        protected override void ManageProduction()
        {
            TimeSpan span = (DateTime.Now - ProductionStart);
            if (span.TotalMinutes + MissionComponent.AiSpeedModifier >= CurrentProductionTime)
            {
                if (!CalculatedOffset)
                {
                    var firewall = MissionComponent.firewall;
                    if (firewall == null)
                    {
                        //no firewall found, go to default, odd that this happened
                        XOffset = 100;
                        YOffset = 0;
                        ZOffset = 0;
                        CalculatedOffset = true;
                    }
                    else
                    {
                        var firewallPos = firewall.CubeGrid.GetPosition();
                        var thispos = Grid.GetPosition();
                        if (firewallPos.X > thispos.X)
                        {
                            XOffset = -100;
                        }
                        else
                        {
                            XOffset = 100;
                        }
                        if (firewallPos.Y > thispos.Y)
                        {
                            YOffset = -100;
                        }
                        else
                        {
                            YOffset = 100;
                        }
                        if (firewallPos.Z > thispos.Z)
                        {
                            ZOffset = -100;
                        }
                        else
                        {
                            ZOffset = 100;
                        }
                        CalculatedOffset = true;
                    }
                }

                IsProducing = false;
                GoOffline();
                Vector3D currentPosition = Grid.GetPosition();
                if (MissionComponent.writeDebug)
                {
                    StringBuilder sb = new StringBuilder();
                    sb.Append("Deploying Large missile on ");
                    sb.Append(currentPosition.X + XOffset);
                    sb.Append(", ");
                    sb.Append(currentPosition.Y + YOffset);
                    sb.Append(", ");
                    sb.Append(currentPosition.Z + YOffset);
                    MissionComponent.Speak("Station", sb.ToString());
                }

                ThreadWrapper.OrderSpawn(ConstructionPrefabName,
                    new Vector3D(currentPosition.X + XOffset, currentPosition.Y + YOffset, currentPosition.Z + ZOffset), Type, this);
            }
        }

        public static void HookAi(IMyEntity ent, int id)
        {
            Missile m = new Missile(ent as IMyCubeGrid, id);
            MissionComponent.AddAiShip(m, id);
        }

        public override int GetBounty()
        {
            return 35;
        }
    }
}
