﻿using System;
using System.CodeDom;
using System.Collections.Generic;
using System.Linq;
using Sandbox.Common.ObjectBuilders;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using VRageMath;

namespace Scripts.KSWH
{
    class Missile : Ai
    {
        private bool Radiation = false;
        private bool doExplode = false;
        private DateTime SpawnTime;
        private static int LiveTime = 5; //minutes

        public static string HudAttackMessage = Hud.MessageLead + "///WARNING::                                                                                                  \n" +
                                                 ">Incomming signature detected\n    -SIZE: tiny\n    -ARMOR: n/a\n    -OFFENSE: n/a\n>HIGH THREAT LEVEL\n\n***IMPACT IMMINENT***";

        public Missile(IMyCubeGrid grid, int id)
        {
            Grid = grid;
            SetId(id);
            //m_cockpit = controller;
            IsControlable = true;
            SetOwsnershipToEnemy(grid);
            IsSpotted = true;
            FindTarget(false);
            MissionComponent.AlarmHandler.MissileWarning();
            MaxCount = 50;
            SpawnTime = DateTime.Now;
        }

        public Missile(IMyCubeGrid grid, IMyControllableEntity controller)
        {
            Grid = grid;
            m_cockpit = controller;
            IsControlable = (controller != null);
            SetOwsnershipToEnemy(grid);
            IsSpotted = true;
            FindTarget(false);
            MissionComponent.AlarmHandler.MissileWarning();
            MaxCount = 50;
            SpawnTime = DateTime.Now;
        }

        public override int GetBounty()
        {
            return 0;
        }

        public override void Update()
        {
            Count++;
            FindTarget(Radiation);

            Target = MissionComponent.firewall.FatBlock.CubeGrid;
            TimeSpan span = (DateTime.Now - SpawnTime);
            if (span.Minutes >= LiveTime)
            {
                SelfDestruct();
            }

            if (!doExplode)
            {
                base.Update();
                if (!IsControlable)
                {
                    return;
                }
            }
            else
            {
                IsControlable = false;
                return;
            }
            

            if (Target == null)
            {
                return; //no target, sleep it off
            }

            var dir = Target.GetPosition() - Grid.GetPosition();

            if (!IsAttacking && dir.Length() < 5000)
            {
                IsAttacking = true;
                MissionComponent.AlarmHandler.Alert((int)Hud.MessagePriority.Missile, HudAttackMessage);
            }

            if (dir.Length() > 100)
            {
                if (nav != null)
                {
                    if (nav.CNS.AutoPilotReady)
                    {
                        var loc = GetTargetDestination(Target.GetPosition(), 0, 0);
                        if (!loc.HasValue)
                        {
                            return;
                        }


                        nav.CNS.setDestination(loc);
                    }
                    return;
                }
                var locb = GetTargetDestination(Target.GetPosition(), 0, 0);
                if (!locb.HasValue)
                {
                    MissionComponent.Speak("ERROR", "nav cannot locate waypoint");
                    return;
                }

                nav = MissionComponent.AutoHandler.AddNavigator(Grid, locb);
                return; 
            }

            if (nav != null)
            {
                MissionComponent.AutoHandler.RemoveNavigator(Grid);
                nav = null;
            }

            var dirNorm = Vector3.Normalize(dir);
            var x = (float) -(m_cockpit as IMyEntity).WorldMatrix.Up.Dot(dirNorm);
            var y = (float) -(m_cockpit as IMyEntity).WorldMatrix.Left.Dot(dirNorm);
            var forw = (m_cockpit as IMyEntity).WorldMatrix.Forward.Dot(dirNorm);

            if (forw < 0)
                y = 0;
            if (Math.Abs(x) < 0.2f)
                x = 0;
            if (Math.Abs(y) < 0.2f)
                y = 0;

            dir = Vector3.TransformNormal(dir, (Matrix) (m_cockpit as IMyEntity).GetWorldMatrixNormalizedInv());

            var rot = new Vector2(x, y)*10f;
            if (dir.LengthSquared() > 0 || rot.LengthSquared() > 0)
            {
                if (dir.Length() < 1500)
                {
                    UseEcm();
                }
                m_cockpit.MoveAndRotate(dir, rot, 0);
                Count = 0;
                

            }
            else
            {
                m_cockpit.MoveAndRotateStopped();
            }
            if (!Radiation && dir.Length() < 200)
            {

                MissionComponent.AlarmHandler.RadWarning();
            }
            if (dir.Length() <= 50)
            {
                doExplode = true;
                SelfDestruct();
            }
        }

        protected override void UseEcm()
        {
            if (!MissionComponent.StationHandler.EcmReady)
            {
                return; //not in incursion state 6
            }

            if (UsedEcm)
            {
                return; //unit already used ecm power
            }

            MissionComponent.EcmHandler.ScrapStorm(EcmActions.Size.Large, Target.GetPosition());
            UsedEcm = true;
        }

        protected override int GetLivespawn()
        {
            return 5;
        }

        private void FindTarget(bool takeDecoy)
        {
            Target = null;
            ThreadWrapper.TargetLock.AcquireExclusive();
            try
            {
                if (takeDecoy)
                {

                    BoundingSphereD range = new BoundingSphereD(Grid.GetPosition(), 200);
                    var ents = MyAPIGateway.Entities.GetEntitiesInSphere(ref range);
                    MyDefinitionId decoyDef = new MyDefinitionId(typeof (MyObjectBuilder_Decoy));
                    foreach (IMyEntity ent in ents)
                    {
                        var grid = ent as IMyCubeGrid;
                        if (grid == null) continue;
                        var lst = new List<IMySlimBlock>();
                        grid.GetBlocks(lst,
                            x => x.FatBlock != null && x.FatBlock.BlockDefinition.TypeId.Equals(decoyDef.TypeId));
                        if (lst.Count <= 0) continue;
                        Target = grid;
                        if (MissionComponent.writeDebug)
                        {
                            MissionComponent.Speak("MIssile", "got a decoy, attacking");
                        }
                        ThreadWrapper.TargetLock.ReleaseExclusive();
                        return;
                    }
                }


                // if (MissionComponent.writeDebug)
                // {
                   // MissionComponent.Speak("Missile", "has target");
                // }
            }
            catch (Exception ex)
            {
                //ignored
            }
           
            ThreadWrapper.TargetLock.ReleaseExclusive();
        }
    }
}
