﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Sandbox.ModAPI;
using VRage;
using VRageMath;

namespace Scripts.KSWH
{
    /// <summary>
    /// Class that will handle radar entities in the game
    /// </summary>
    class RadarHandler
    {
        private FastResourceLock radarLock;
        private List<BoundingSphere> coverage; 
        private static int radarRange = 5000;
        public bool hasRadar { private set; get; }

        /// <summary>
        /// default constructor
        /// </summary>
        public RadarHandler()
        {
            coverage = new List<BoundingSphere>();
            radarLock = new FastResourceLock();
        }

        /// <summary>
        /// Scan the game for radars and add them to the list
        /// </summary>
        public void ReScanRadars()
        {
            radarLock.AcquireExclusive();
            hasRadar = false;
            coverage.Clear();
            HashSet<IMyEntity> entities = new HashSet<IMyEntity>();
            MyAPIGateway.Entities.GetEntities(entities);
            foreach (IMyEntity ent in entities)
            {
                List<IMySlimBlock> blocks = new List<IMySlimBlock>();
                var grid = ent as IMyCubeGrid;
                if (grid == null)
                {
                    continue;
                }
                grid.GetBlocks(blocks, x => x.FatBlock != null && x.FatBlock.BlockDefinition.SubtypeId.Equals("RadarLarge"));
                if (blocks.Count == 0)
                {
                    continue;
                }
                var tmpList = blocks.ToList();
                foreach (IMySlimBlock radar in tmpList)
                {
                    coverage.Add(new BoundingSphere(radar.FatBlock.GetPosition(), radarRange));
                }
                
            }
            hasRadar = coverage.Count != 0;
            radarLock.ReleaseExclusive();
        }

        public bool InRange(Vector3D ent)
        {
            List<BoundingSphere> ranges = new List<BoundingSphere>();
            radarLock.AcquireExclusive();
            ranges.AddRange(coverage);
            radarLock.ReleaseExclusive();
            return ranges.Select(range => range.Contains(ent)).Any(res => res != ContainmentType.Disjoint);
        }
    }
}
