﻿using System;

namespace Scripts.KSWH
{
    class Randomizer
    {
        private UInt32 salt = 0; //simple little thing to reduce the change of a duplication
        private float part;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="max">The maximum bound of the int, NOTE: must be between 1 and Int.maxvalue</param>
        public Randomizer(int max)
        {
            if (max < 0) {
                return;
            }
            part = UInt32.MaxValue / max;
        }

        /// <summary>
        /// Create a simple random (low change of it bieng unique)
        /// </summary>
        /// <param name="data">byte array with a root</param>
        /// <returns>a randomized int (not always unique)</returns>
        public double Hash(Byte[] data)
        {
            ++salt;
            return Hash(data, 0xc58f1a7b);
        }
        
        private const UInt32 m = 0x5bd1e995; //black magic seeds
        private const Int32 r = 24; //more black magic 

        /// <summary>
        /// Create a random integer value with a bigger change of bieng unique
        /// </summary>
        /// <param name="data">byte array with a root</param>
        /// <param name="seed">seed number</param>
        /// <returns>unique int</returns>
        public double Hash(Byte[] data, UInt32 seed)
        {
            seed = seed + salt;
            ++salt;
            Int32 length = data.Length;
            if (length == 0)
                return 0;
            UInt32 h = seed ^ (UInt32)length;
            Int32 currentIndex = 0;
            while (length >= 4)
            {
                UInt32 k = (UInt32)(data[currentIndex++] | data[currentIndex++] << 8 | data[currentIndex++] << 16 | data[currentIndex++] << 24);
                k *= m;
                k ^= k >> r;
                k *= m;

                h *= m;
                h ^= k;
                length -= 4;
            }
            switch (length)
            {
                case 3:
                    h ^= (UInt16)(data[currentIndex++] | data[currentIndex++] << 8);
                    h ^= (UInt32)(data[currentIndex] << 16);
                    h *= m;
                    break;
                case 2:
                    h ^= (UInt16)(data[currentIndex++] | data[currentIndex] << 8);
                    h *= m;
                    break;
                case 1:
                    h ^= data[currentIndex];
                    h *= m;
                    break;
                default:
                    MissionComponent.Speak("System", "God has touched your CPU and created a non binary bit, please call the pope now");
                    break;
            }

            // Do a few final mixes of the hash to ensure the last few
            // bytes are well-incorporated.

            h ^= h >> 13;
            h *= m;
            h ^= h >> 15;

            return h / part;
        }

        /// <summary>
        /// Reset the salt, has to be done when a INT32 goes past its maximum value
        /// </summary>
        public void ResetSalt()
        {
            salt = 0;
        }
    }
}
