﻿using System;
using Sandbox.ModAPI;
using VRageMath;

namespace Scripts.KSWH
{
    class SmallFactoryStation : Station
    {
        private const int MinDistance = 20;
        private const int ProductionTime = 1;
        private int ResourceBonus = 25;
        public const int BuildCosts = 500;

        public SmallFactoryStation()
        {
            Random r = new Random();
            Type = StationType.SmallLogistic;
            ThreadWrapper.OrderSpawn(StaticInformation.SmallLogisticStations[r.Next(StaticInformation.SmallLogisticStations.Length - 1)],
                    GenerateLocation(MinDistance), this);
            Income = ResourceBonus;
            SetOwsnershipToEnemy();
        }

        public SmallFactoryStation(IMyCubeGrid grid)
        {
            Grid = grid;
            Type = StationType.SmallLogistic;
            Income = ResourceBonus;
            SetOwsnershipToEnemy();
        }

        protected override void ManageResources()
        {
            base.ManageResources();
            if (ResourceBonus < 200)
            {
                ResourceBonus++;
                Income = ResourceBonus;
            }
        }

        

        public void StartProduction()
        {
            if (!IsProducing)
            {
                if (!MissionComponent.host.AiPay(40))
                {
                    return;
                }
                Random r = new Random(DateTime.Now.Minute);
                string prefab = StaticInformation.SmallMissiles[r.Next(StaticInformation.SmallMissiles.Length - 1)];
                base.StartProduction(ProductionTime, prefab);
                if (MissionComponent.writeDebug)
                {
                    MissionComponent.Speak("Small launcher", "Now building missile");
                }
            }
        }

        public override int GetBounty()
        {
            return 10;
        }
    }
}
