﻿using System;
using System.Collections.Generic;
//using System.Linq;
using System.Text;

//using Sandbox.Common;
using Sandbox.Definitions;
using Sandbox.ModAPI;
using VRage;

namespace Scripts.KSWH
{
	/// <summary>
	/// Generates files to be read by GamutLogViewer
	/// using Log4J Pattern: [%date][%level][%Grid][%Class][%Method][%PriState][%SecState]%Message
	/// </summary>
	[Sandbox.Common.MySessionComponentDescriptor(Sandbox.Common.MyUpdateOrder.NoUpdate)]
	public class Logger : Sandbox.Common.MySessionComponentBase
	{
		private static System.IO.TextWriter logWriter = null;
		private StringBuilder stringCache = new StringBuilder();

		private static int maxNumLines = int.MaxValue;
		private static int numLines = 0;

		private string gridName, className;

		public Logger() { }

		public Logger(string gridName, string className)
		{
			this.gridName = gridName;
			this.className = className;

			if (logWriter == null)
				logWriter = MyAPIGateway.Utilities.WriteFileInLocalStorage("Sector15AIPilotLog.txt", typeof(Logger)); // was getting a NullReferenceException in ..cctor(), so we delay
		}

		private static FastResourceLock lock_log = new FastResourceLock();

		public void log(string toLog, string methodName, severity level, string primaryState = null, string secondaryState = null) { log(level, methodName, toLog, primaryState, secondaryState); }

		public void log(severity level, string methodName, string toLog, string primaryState = null, string secondaryState = null)
		{
			if (logWriter == null)// || !canLog(level))
				return;

			if (toLog == null)
				toLog = "no message";
			if (numLines >= maxNumLines)
			{
				numLines = 0;
				new Logger(null, "Logger").log(severity.INFO, "log()", "reached maximum log length");
				//minSeverity = severity.OFF;
				close();
				return;
			}

			lock_log.AcquireExclusive();
			try
			{
				numLines++;
				appendWithBrackets(DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss,fff"));
				appendWithBrackets(level.ToString());
				appendWithBrackets(gridName);
				appendWithBrackets(className);
				appendWithBrackets(methodName);
				appendWithBrackets(primaryState);
				appendWithBrackets(secondaryState);
				stringCache.Append(toLog);

				logWriter.WriteLine(stringCache);
				logWriter.Flush();
				stringCache.Clear();
			}
			finally
			{
				lock_log.ReleaseExclusive();
			}
		}

		private void appendWithBrackets(string append)
		{
			if (append == null)
				append = String.Empty;
			stringCache.Append('[');
			stringCache.Append(append);
			stringCache.Append(']');
		}

		/// <summary>
		/// closes the static log file
		/// </summary>
		private void close()
		{
			if (logWriter == null)
				return;
			logWriter.Close();
			logWriter = null;
		}

		//public override void UpdateBeforeSimulation()
		//{
		//}

		protected override void UnloadData()
		{ close(); }

		public enum severity : byte { OFF, FATAL, ERROR, WARNING, INFO, DEBUG, TRACE, ALL}
		//private static severity minSeverity = severity.ALL;
		//public static bool canLog(severity level)
		//{ return (level.CompareTo(minSeverity) <= 0); }
	}
}
