// line removed by build.py 
using System;
using System.Collections.Generic;
//using System.Linq;
using System.Text;

using Sandbox.Common;
using Sandbox.ModAPI;

namespace Scripts.KSWH
{
	internal static class Settings
	{
		// Settings
		public enum BoolSetName : byte { bAllowJumpCreative, bAllowJumpSurvival, bAllowAutopilot }
		public enum IntSetName : byte { }
		public enum FloatSetName : byte { fDefaultSpeed }

		public static Dictionary<BoolSetName, bool> boolSettings = new Dictionary<BoolSetName, bool>();
		public static Dictionary<IntSetName, int> intSettings = new Dictionary<IntSetName, int>();
		public static Dictionary<FloatSetName, float> floatSettings = new Dictionary<FloatSetName, float>();

		//
		private static string settings_file_name = "AutopilotSettings.txt";
		private static System.IO.TextReader settingsReader;
		private static System.IO.TextWriter settingsWriter;

		private static string strVersion = "Version";
		private static int latestVersion = 17; // in sequence of updates on steam

		private static Logger myLogger = new Logger(null, "Settings");
		[System.Diagnostics.Conditional("LOG_ENABLED")]
		private static void log(string toLog, string method = null, Logger.severity level = Logger.severity.DEBUG)
		{ myLogger.log(level, method, toLog); }
		
		private static void buildSettings()
		{
			boolSettings.Add(BoolSetName.bAllowJumpCreative, true);
			boolSettings.Add(BoolSetName.bAllowJumpSurvival, true);
			boolSettings.Add(BoolSetName.bAllowAutopilot, true);
			//boolSettings.Add(BoolSetName.bAllowMessaging, true);
			//boolSettings.Add(BoolSetName.bAllowRadar, true);
			floatSettings.Add(FloatSetName.fDefaultSpeed, 100);
			//floatSettings.Add(FloatSetName.fMinimumGridVolumeLarge, 100);
			//floatSettings.Add(FloatSetName.fMinimumGridVolumeSmall, 10);
		}

		internal static void Open()
		{
		    boolSettings.Add(BoolSetName.bAllowJumpCreative, false);
            boolSettings.Add(BoolSetName.bAllowJumpSurvival, false);
            boolSettings.Add(BoolSetName.bAllowAutopilot, true);
            floatSettings.Add(FloatSetName.fDefaultSpeed, 100);
            

			/*buildSettings();

			int fileVersion = readAll();
			if (fileVersion != latestVersion)
				MyAPIGateway.Utilities.ShowNotification("Autopilot has been updated.", 10000, MyFontEnum.White);
			log("file version: "+fileVersion+", latest version: "+latestVersion);
*/
			//readAll();
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns>version of file</returns>
		private static int readAll()
		{
			if (!MyAPIGateway.Utilities.FileExistsInLocalStorage(settings_file_name, typeof(Settings)))
				return -1;
			try
			{
				settingsReader = MyAPIGateway.Utilities.ReadFileInLocalStorage(settings_file_name, typeof(Settings));

				string[] versionLine = settingsReader.ReadLine().Split('=');
				if (versionLine.Length != 2 || !versionLine[0].Equals(strVersion))
					return -2; // first line is not version
				int fileVersion;
				if (!int.TryParse(versionLine[1], out fileVersion))
					return -3; // could not parse version

				// read settings
				while (true)
				{
					string line = settingsReader.ReadLine();
					//log("got a line: " + line, "readAll()", Logger.severity.TRACE);
					if (line == null)
						break;
					parse(line);
				}

				return fileVersion;
			}
			finally
			{
				settingsReader.Close();
				settingsReader = null;
			}
		}

		internal static void writeAll()
		{
            //TODO: no longer will we write settings, clean this code up in the future

			/*// write to file
			settingsWriter = MyAPIGateway.Utilities.WriteFileInLocalStorage(settings_file_name, typeof(Settings));

			write(strVersion, latestVersion.ToString()); // must be first line

			// write settings
			foreach (KeyValuePair<BoolSetName, bool> pair in boolSettings)
				write(pair.Key.ToString(), pair.Value.ToString());
			foreach (KeyValuePair<IntSetName, int> pair in intSettings)
				write(pair.Key.ToString(), pair.Value.ToString());
			foreach (KeyValuePair<FloatSetName, float> pair in floatSettings)
				write(pair.Key.ToString(), pair.Value.ToString());

			settingsWriter.Flush();
			settingsWriter.Close();
			settingsWriter = null;*/
		}

		// format shall be Name=value
		private static void write(string name, string value)
		{
			StringBuilder toWrite = new StringBuilder();
			toWrite.Append(name);
			toWrite.Append('=');
			toWrite.Append(value);
			settingsWriter.WriteLine(toWrite);
		}

		private static void parse(string line)
		{
			if (line.Length < 4)
			{
				log("line too short: " + line, "parse()", Logger.severity.WARNING);
				return;
			}

			char first = line[0];
			string[] split = line.Split('=');

			if (split.Length != 2)
			{
				log("split wrong: " + split.Length, "parse()", Logger.severity.WARNING);
				return;
			}
			
			string name = split[0];
			string value = split[1];

			switch (first)
			{
				case 'b':
					{
						BoolSetName bsn;
						bool b;
						if (!Enum.TryParse<BoolSetName>(name, out bsn) || !bool.TryParse(value, out b))
							return;
						log("got bool setting: " + bsn + ", with value: " + b, "parse()", Logger.severity.TRACE);
						boolSettings[bsn] = b;
						return;
					}
				case 'i':
					{
						IntSetName isn;
						int i;
						if (!Enum.TryParse<IntSetName>(name, out isn) || !int.TryParse(value, out i))
							return;
						log("got int setting: " + isn + ", with value: " + i, "parse()", Logger.severity.TRACE);
						intSettings[isn] = i;
						return;
					}
				case 'f':
					{
						FloatSetName fsn;
						float f;
						if (!Enum.TryParse<FloatSetName>(name, out fsn) || !float.TryParse(value, out f))
							return;
						floatSettings[fsn] = f;
						log("got float setting: " + fsn + ", with value: " + f, "parse()", Logger.severity.TRACE);
						return;
					}
			}
			log("failed to parse: " + line, "parse()", Logger.severity.WARNING);
		}
	}
}
